/*

Highlands Electronics Simple Disk Operating System

*/
#include  "device_430.h"
#include "HEDisk.h"

//basic succsess is 'GetSectorData for now...
//The MSP430 and x86 are little endian, so nothing special
bool HEDsk_LoadVolumeInfo( )
{
	//can't be zero to start cause of caching
	heVol.curSector= 0xffff;
	
	//Get Master Boot Record (MBR), sector 0
	if( ! GetSectorData( 0 ) )
		return false;

	//Get the sector of the first partition from the MBR
	heVol.bootSector= *(DWORD*)( &secBuf[ MBR_LBA0 ] );

	//Get the FAT Boot Sector
	if( ! GetSectorData( heVol.bootSector ) )
		return false;

	//The root sector of the FAT table
	heVol.rootFat= heVol.bootSector + *(WORD*)( &secBuf[ HEFAT_RESERVED ] );

	//Root sector of the directory
	heVol.rootDir= heVol.rootFat + ( *(WORD*)( &secBuf[ HEFAT_SECPERFAT ] ) << 1 );

	//count of directory entries
	heVol.dirEntries= *(WORD*)( &secBuf[ HEFAT_DIRENTRIES ] );

	//Sectors per cluster
	heVol.secPerClust= secBuf[ HEFAT_SECPERCLST ];

	//Bytes per cluster for seek
	heVol.bytesPerClust= heVol.secPerClust << HEDS_SECTOR_SHIFT;

	//Root sector of the data area
	heVol.rootData= heVol.rootDir + ( heVol.dirEntries >> 3 ) + 1;

	//Set the first free fat position to start
	// not really free, need to call HEDsk_GetFreeFat(...)
	//Fragmentation is not our concern in this app
	heVol.freeFatPos= 0x0002;

	//start with flags reset
	heVol.volFlags= 0;
	//and current sector
	heVol.curSector= 0;

	return true;
}

bool HEDsk_GetFileInfo( pHEDskFileInfo pInfo, char* lpsCanonicalName )
{
	//Get the root dir sector to start search for file
	GetSectorData( heVol.rootDir );

	long dirCount= 1;
	long dirSecCount= heVol.rootDir;
	BYTE* pEnt= secBuf;
	for( ; dirCount < heVol.dirEntries; pEnt+= HEFAT_SZF_DIRENT )
	{
		if( ! ( ++dirCount & HEDS_DIRCNTMASK ) )
		{
			//need to move to next sector and reset pEnt;
			GetSectorData( ++dirSecCount );
			pEnt= secBuf;
		}
		if( ( pEnt[ 11 ] & 0x1f ) || *pEnt == 0xe5 )
			//We are just looking for real cananonical file names
			continue;

		if( acstrcmp( lpsCanonicalName, (char*)pEnt, HEFAT_SZF_FNAME ) )
		{
			//found it
			pInfo->dirPos= (WORD)( pEnt - secBuf );
			pInfo->dirSect= (WORD)dirSecCount;
			//set file pointer to begining
			pInfo->curClust= *(WORD*)( &pEnt[ HEDS_FILE_DIR_CLUST ] );
			pInfo->size= *(DWORD*)( &pEnt[ HEDS_FILE_SIZE ] );
			pInfo->curPointer= 0;
			pInfo->curSectorPos= 0;
			return true;
		}
	}
	return false;
}

void HEDsk_GetFreeFat( )
{
	//need sector and pos.
	DWORD fatSec= (WORD)heVol.rootFat + ( heVol.freeFatPos >> HEDS_FATSECSHIFT );
	BYTE fatPos= heVol.freeFatPos & HEDS_HALFSECMASK;

	for( ; ; )
	{
		//This caches so don't check refetch
		GetSectorData( fatSec );

		//Is cluster free?
		if( ! *(WORD*)&secBuf[ fatPos << 1 ] )
			break;

		if( ! ( ++fatPos & HEDS_ENDFATMASK ) )
		{
			//move to start of next fat sector
			fatPos= 0;
			++fatSec;
			//TODO ? check if partition is full ?!
		}
	}
	//save as a FAT entry
	heVol.freeFatPos= (WORD)( ( fatSec - heVol.rootFat ) << HEDS_FATSECSHIFT ) | fatPos;
	//mark it as an end cluster, only reason for finding it
	*(WORD*)&secBuf[ fatPos << 1 ]= HEDS_ENDCLUSTMSK;
	//Mark sector dirty
	SetSectorData( fatSec );
}

void HEDsk_AddCluster( pHEDskFileInfo pInfo )
{
	HEDsk_GetFreeFat( );

	//Is the file zero length?
	if( ! pInfo->curClust )
	{
		GetSectorData( pInfo->dirSect );
		//I've seen this non zero when it should be. Just me?
		//*(DWORD*)&secBuf[ pInfo->dirPos + HEDS_FILE_SIZE ]= 0;
		*(WORD*)&secBuf[ pInfo->dirPos + HEDS_FILE_DIR_CLUST ]= heVol.freeFatPos;
		SetSectorData( pInfo->dirSect );
	}
	else
	{
		DWORD curSector= ( pInfo->curClust >> HEDS_FATSECSHIFT )  + heVol.rootFat;
		GetSectorData( curSector );

		//update the last cluster entry in the FAT
		*(WORD*)&secBuf[ ( pInfo->curClust & HEDS_HALFSECMASK ) << 1 ]= heVol.freeFatPos;
		SetSectorData( curSector );
	}
	//Update the file info
	pInfo->curClust= heVol.freeFatPos;
}

void HEDsk_SeekFile( pHEDskFileInfo pInfo, DWORD pos )
{
	//This is the byte count offset into the current file position
	DWORD absPos= pInfo->curPointer + ( (DWORD)pInfo->curSectorPos << HEDS_SECTOR_SHIFT );

	if( absPos > pos )
	{
		//there is a start pos with cur cluster that pos may be greater than, but
		//rewind and start over
		absPos= 0;
		pInfo->curSectorPos= 0;
		GetSectorData( pInfo->dirSect );
		pInfo->curClust= *(WORD*)&secBuf[ pInfo->dirPos + HEDS_FILE_DIR_CLUST ];
	}

	//subPos is byte to go
	DWORD subPos= pos - absPos;

	//first is to move by cluster
	for( ; subPos > heVol.bytesPerClust; )
	{
		DWORD fsector= ( pInfo->curClust >> HEDS_FATSECSHIFT ) + heVol.rootFat;
		GetSectorData( fsector );
		pInfo->curClust= *(WORD*)( &secBuf[ ( pInfo->curClust & HEDS_SECTORMASK ) << 1 ] );

		pInfo->curSectorPos+= heVol.secPerClust;
		subPos-= heVol.bytesPerClust;
		continue;
	}
	//At this point we should be in the cluster
	pInfo->curSectorPos+= subPos >> HEDS_SECTOR_SHIFT;
	pInfo->curPointer= (WORD)pos & HEDS_SECTORMASK;
}

bool HEDsk_ReadFile( pHEDskFileInfo pInfo, BYTE* buf, WORD count )
{
	//Absolute sector is not kept, should it?
	DWORD curSector= HEDsk_GetCurFileSector( pInfo );

	//out data pos
	BYTE* pOutBuf= buf;

	for( ; count; )
	{
		WORD readSize;

		if( pInfo->curPointer + count > HEDS_SECTORSIZE )
			readSize= HEDS_SECTORSIZE - (WORD)pInfo->curPointer;
		else
			readSize= count;

		GetSectorData( curSector );
		HEDsk_memcpy( pOutBuf, secBuf + pInfo->curPointer, readSize );

		pInfo->curPointer+= readSize;
		if( pInfo->curPointer == HEDS_SECTORSIZE )
			pInfo->curPointer= 0;

		pOutBuf+= readSize;
		count-= readSize;

		//Need more sectors?
		if( ! count )
			break;

		++curSector
		//move into the next cluster?
		;
		if( ! ( ++pInfo->curSectorPos & ( heVol.secPerClust - 1 ) ) )
		{
			DWORD fsector= ( pInfo->curClust >> HEDS_FATSECSHIFT ) + heVol.rootFat;
			GetSectorData( fsector );
			pInfo->curClust= *(WORD*)( &secBuf[ ( pInfo->curClust & HEDS_SECTORMASK ) << 1 ] );
			curSector= HEDsk_GetCurFileSector( pInfo );
		}
	}
	return true;
}

bool HEDsk_WriteFile( pHEDskFileInfo pInfo, BYTE* buf, WORD count )
{
	//cause...
	if( ! count )
		return true;

	//special case, file length is zero. Don't trust 'size' ?
	if( ! pInfo->curClust )
		HEDsk_AddCluster( pInfo );

	//Absolute sector is not kept, should it?
	DWORD curSector= HEDsk_GetCurFileSector( pInfo );

	//If we were at the end of the sector already, this didn't need to happen
	GetSectorData( curSector );

	//working counter, need count later for dir
	WORD workCount= count;

	//in data pos
	BYTE* pInBuf= buf;

	for( ; workCount; )
	{
		WORD writeSize;
		if( pInfo->curPointer + workCount > HEDS_SECTORSIZE )
			writeSize= HEDS_SECTORSIZE - (WORD)pInfo->curPointer;
		else
			writeSize= workCount;
		
		HEDsk_memcpy( secBuf + pInfo->curPointer, pInBuf, writeSize );
		pInfo->curPointer+= writeSize;
		SetSectorData( curSector );

		workCount-= writeSize;

		//Need more sectors?
		if( workCount )
		{
			//bump curSectorPos, move into the next cluster?
			if( ! ( ++pInfo->curSectorPos & ( heVol.secPerClust - 1 ) ) )
			{
				//Add the cluster
				HEDsk_AddCluster( pInfo );

				//Need the data sector back for subsequent writes
				curSector= HEDsk_GetCurFileSector( pInfo );
			}
			else
				//just move up through cluster
				++curSector;

			//freshen up for partial writes
			GetSectorData( curSector );
			pInBuf+= writeSize;
			pInfo->curPointer= 0;
		}
	}
	pInfo->size+= count;
	return true;
}

bool HEDsk_FlushFile( pHEDskFileInfo pInfo )
{
	GetSectorData( pInfo->dirSect );
	*(DWORD*)&secBuf[ pInfo->dirPos + HEDS_FILE_SIZE ]= pInfo->size;
	SetSectorData( pInfo->dirSect );
	GetSectorData( 0 );
	return true;
}

DWORD HEDsk_GetCurFileSector( pHEDskFileInfo pInfo )
{
	return
		heVol.rootData
		+ ( pInfo->curClust - 2 ) * (WORD)heVol.secPerClust //this should be shift, not call to mult(..)
		+ ( pInfo->curSectorPos & ( heVol.secPerClust - 1 ) );
}


