
//#include "stdafx.h"
#define _WIN32_WINNT 0x0501
#include <windows.h>
#include <stdio.h>

#define BUFSIZE 512
#define FILESYSNAMEBUFSIZE 512

BOOL ProcessVolumeMountPoint (HANDLE hPt,
                              char *PtBuf, DWORD dwPtBufSize,
                              char *Buf, DWORD dwBufSize)
{
   BOOL bFlag;                   // Boolean result   
   char Path[BUFSIZE];           // construct a complete path here
   HANDLE hFf, hFile;            // file search, file handles
   WIN32_FIND_DATA fd;           // find data for a file search
   DWORD BytesReturned;          // from DeviceIoControl call
   char Target[BUFSIZE];         // target of mount at mount point

   printf ("\tVolume mount point found is \"%s\"\n", PtBuf);

   // Detect the volume mounted there. Unfortunately, there is no 
   // simple way to map a unique volume name to a DOS drive letter. 
   // One way to do that is to build an associative array, which we 
   // will "leave as an exercise for the student." 

   // Build a unique path to the mount point

   strcpy (Path, Buf);
   strcat (Path, PtBuf);

   bFlag = GetVolumeNameForVolumeMountPoint(
                Path, // input volume mount point or directory
              Target, // output volume name buffer
              BUFSIZE // size of volume name buffer
           );

   if (!bFlag)
      printf ("\tAttempt to get volume name for %s failed.\n", Path);
   else 
      printf ("\tTarget of the volume mount point is %s.\n", Target);
  
   // Now, either get the next mount point and return it, or return a 
   // value indicating there are no more mount points.

   bFlag = FindNextVolumeMountPoint(
                     hPt, // handle to scan
                   PtBuf, // pointer to output string
              dwPtBufSize // size of output buffer
           );
  return (bFlag);
}

// Process each volume. This makes maintenance easier than doing it
// in line. The Boolean result indicates whether there is another 
// volume to be scanned. 

BOOL ProcessVolume (HANDLE hVol, char *Buf, int iBufSize)
{
   BOOL bFlag;                  // generic results flag for return
   HANDLE hPt;                  // handle for mount point scan
   char PtBuf[BUFSIZE];         // string buffer for mount points
   DWORD dwSysFlags;            // flags that describe the file system
   char FileSysNameBuf[FILESYSNAMEBUFSIZE];

   printf ("Volume found is \"%s\".\n", Buf);

   // Is this volume NTFS? 
   GetVolumeInformation( Buf, NULL, 0, NULL, NULL,
                         &dwSysFlags, FileSysNameBuf, 
                         FILESYSNAMEBUFSIZE);

   // Detect support for reparse points, and therefore for volume 
   // mount points, which are implemented using reparse points.

   if (! (dwSysFlags & FILE_SUPPORTS_REPARSE_POINTS)) 
   {
      printf ("\tThis file system does not support volume mount points.\n");
   } 
   else 
   {
      // Start processing mount points on this volume. 

      hPt = FindFirstVolumeMountPoint(
                  Buf, // root path of volume to be scanned
                PtBuf, // pointer to output string
               BUFSIZE // size of output buffer
            );

      if (hPt == INVALID_HANDLE_VALUE)
      {
         printf ("\tNo volume mount points found!\n");
      } 
      else 
      {
         // Process the volume mount point.
         bFlag = ProcessVolumeMountPoint (hPt, PtBuf, BUFSIZE, Buf, BUFSIZE); 

         // Do while we have volume mount points to process.
         while (bFlag) 
            bFlag = ProcessVolumeMountPoint (hPt, PtBuf, BUFSIZE, Buf, BUFSIZE);

         FindVolumeMountPointClose(hPt);
       }
   }

  // Stop processing mount points on this volume. 

   bFlag = FindNextVolume(
                hVol, // handle to scan being conducted
                 Buf, // pointer to output
             iBufSize // size of output buffer
           );

   return (bFlag); 
}

/*
int main(void)
{
   char buf[BUFSIZE];            // buffer for unique volume identifiers
   HANDLE hVol;                  // handle for the volume scan
   BOOL bFlag;                   // generic results flag

   // Open a scan for volumes.
   hVol = FindFirstVolume (buf, BUFSIZE );

   if (hVol == INVALID_HANDLE_VALUE)
   {
      printf ("No volumes found!\n");
      return (-1);
   }

   // We have a volume; process it.
   bFlag = ProcessVolume (hVol, buf, BUFSIZE);

   // Do while we have volumes to process.
   while (bFlag) 
   {
      bFlag = ProcessVolume (hVol, buf, BUFSIZE);
   }

   // Close out the volume scan.
   bFlag = FindVolumeClose(
              hVol  // handle to be closed
           );

   return (bFlag);
}
*/


bool GetSector( BYTE* buf )
{
	HANDLE hDriver;
	char nb[ 512 ];

	DWORD dw;
	DWORD dw2;
	DWORD dw3;
	char tBuf[ 512 ];
	char tBuf2[ 512 ];
	char nameBuf[ 512 ];

	BOOL bTest= GetLogicalDriveStrings( 512, nb );

	DWORD ser;
	DWORD maxL;

	BOOL bDInfo= GetVolumeInformation( "g:\\", tBuf, 512, &dw, &dw2, &dw3, nb, 512 );
	//Return Values
	//If the function succeeds, the return value is a bitmask representing the currently available disk drives. Bit position 0 (the least-significant bit) is drive A, bit position 1 is drive B, bit position 2 is drive C, and so on.
	DWORD drives= GetLogicalDrives( );

	hDriver= FindFirstVolume( (LPSTR)buf, 512 );
	if( hDriver == INVALID_HANDLE_VALUE )
		return false;

	for( ; ; )
	{
		//for( ; ; )
		//{
		//	DWORD dwSysFlags;            // flags that describe the file system
		//	BOOL bVI= GetVolumeInformation( (LPSTR)buf, nameBuf, 512, &ser, &maxL, &dwSysFlags, (LPSTR)tBuf2, 512 );
		//	TRACE(" Vol_name: %s\n", buf );

		//	if( !! ( dwSysFlags & FILE_SUPPORTS_REPARSE_POINTS ) )
		//	{
		//		HANDLE hFirstVolPt= FindFirstVolumeMountPoint(
		//			(LPSTR)buf, // root path of volume to be scanned
		//			tBuf2, // pointer to output string
		//			512 // size of output buffer
		//		);
		//		if( ! ( hFirstVolPt == INVALID_HANDLE_VALUE ) )
		//		{
		//			BOOL bFlag= ProcessVolumeMountPoint( hFirstVolPt, (LPSTR)buf, 512, nameBuf, 512 ); 

		//			// Do while we have volume mount points to process.
		//		//	while (bFlag) 
		//		//bFlag = ProcessVolumeMountPoint (hPt, PtBuf, BUFSIZE, Buf, BUFSIZE);

		//			FindVolumeMountPointClose( hFirstVolPt );
		//		}
		//		else
		//		{
		//			DWORD error= GetLastError( );
		//			TRACE( "FindFirstVolumeMountPoint error: %d\n", error );
		//			break;
		//		}
		//	}
		//	else
		//	{
		//		TRACE( "this folume does not have mount points: %s\n", buf );
		//		break;

		//	}
		//}//for volumes points

		//remvove trailing /
		CString strDName( "\\\\.\\\\c:" );
		HANDLE hDevice2= CreateFile( strDName,
			GENERIC_READ | GENERIC_WRITE,
			FILE_SHARE_READ | FILE_SHARE_WRITE,
			0,
			OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL,
			0
		);

		HANDLE hDevice= CreateFile( (LPSTR)buf,
			GENERIC_READ | GENERIC_WRITE,
			FILE_SHARE_READ | FILE_SHARE_WRITE,
			0,
			OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL,
		);

		//FSCTL_GET_NTFS_VOLUME_DATA
		DWORD dwIn;
		NTFS_VOLUME_DATA_BUFFER ntfsBuf;

		BOOL bDIO= DeviceIoControl( hDevice, FSCTL_GET_NTFS_VOLUME_DATA,
			NULL, //not used
			0, //not used
			&ntfsBuf,
			sizeof( NTFS_VOLUME_DATA_BUFFER ),
			&dwIn,
			NULL
		);

		BOOL bFlag= FindNextVolume(
			hDriver, // handle to scan being conducted
			(LPSTR)buf, // pointer to output
			512 // size of output buffer
		);
		if( ! bFlag )
			break;

	}//for volumes

	//GUID_DEVCLASS_CDROM
	GetVolumeNameForVolumeMountPoint( "G:\\", nb, 512 );

//HANDLE hDriver= CreateFile( "\\.\G:",
	hDriver= CreateFile( "\\\\.\\PhysicalDrive0",
	GENERIC_READ | GENERIC_WRITE,
	FILE_SHARE_READ | FILE_SHARE_WRITE,
	0,
	OPEN_EXISTING,
	FILE_ATTRIBUTE_NORMAL,
		0);

	DWORD error= GetLastError( );
	
	DWORD byteIn;
	BOOL test= ReadFile( hDriver, aSector, 512, &byteIn, 0 );

	TRACE( "test: %d\n", test );

	return true;
}

